---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ada.Exceptions;

with Ast2Cfg.Utils; use Ast2Cfg.Utils;
with Ast2Cfg.CFGs;
with Ast2Cfg.Pkgs;
with Ast2Cfg.Nodes;
with Ast2Cfg.Output; use Ast2Cfg.Output;

package body Ast2Cfg.Flow_Types is

   procedure Init (Flow: in out Flow_Object;
                   Oid:  in Id.Pointer;
                   Name: in Unbounded_String;
                   Pred: in Flow_Class_Ptr := null;
                   Element: in Asis.Element := Asis.Nil_Element) is
   begin
      Flow.Oid := Oid;
      Flow.Name := Name;
      Flow.PTree := null;
      Flow.Element := Element;
      Flow_Class_Ptr_List.Init (Flow.Preds);
      Flow_Class_Ptr_List.Init (Flow.Succs);
      if Pred /= null then
         Add_Pred(Flow, Pred);
      end if;
      Declaration_Ptr_List.Init (Flow.Decls);
      Formal_Decl_Class_Ptr_List.Init (Flow.FDecls);
      Alias_Class_Ptr_List.Init (Flow.Rens);
      Ref_Class_Ptr_List.Init (Flow.Refs);
      Type_Descr_Ptr_List.Init (Flow.Types);
   end Init;

   procedure Free (Flow: in out Flow_Class_Ptr) is
      procedure Destroy_Pkg is new Ada.Unchecked_Deallocation(Pkgs.Pkg_Object'Class, Pkgs.Pkg_Class_Ptr);
      procedure Destroy_Cfg is new Ada.Unchecked_Deallocation(CFGs.CFG_Object'Class, CFGs.CFG_Class_Ptr);
      procedure Destroy_Node is new Ada.Unchecked_Deallocation(Nodes.Node_Object'Class, Nodes.Node_Class_Ptr);
   begin
      if Flow /= null then
         Id.Free(Flow.Oid);
         Nodes.Free_Tree_Recursive(Nodes.Node_Class_Ptr(Flow.PTree));
         Flow_Class_Ptr_List.Clear(Flow.Preds);
         Flow_Class_Ptr_List.Clear(Flow.Succs);
         Declaration_Ptr_List.Free(Flow.Decls);
         Formal_Decl_Class_Ptr_List.Free(Flow.FDecls);
         Alias_Class_Ptr_List.Free(Flow.Rens);
         Ref_Class_Ptr_List.Free(Flow.Refs);
         Type_Descr_Ptr_List.Free(Flow.Types);

         -- I know this is ugly, however, Ada.Finalization turned out to be
         -- unusable with this program
         if Flow.all in Pkgs.Pkg_Object'Class then
            Pkgs.Free_Special(Pkgs.Pkg_Class_Ptr(Flow));
            declare
            begin
               Destroy_Pkg(Pkgs.Pkg_Class_Ptr(Flow));
            exception
               when Ex : others =>
                  Put_Line(Warning, "[Flow_Types] " & Ada.Exceptions.Exception_Name(Ex)  & ": Deallocation of a Pkg failed.");
            end;
         elsif Flow.all in CFGs.CFG_Object'Class then
            CFGs.Free_Special(CFGs.CFG_Class_Ptr(Flow));
            declare
            begin
               Destroy_Cfg(CFGs.CFG_Class_Ptr(Flow));
            exception
               when Ex : others =>
                  Put_Line(Warning, "[Flow_Types] " & Ada.Exceptions.Exception_Name(Ex)  & ": Deallocation of a CFG failed.");
            end;
         elsif Flow.all in Nodes.Node_Object'Class then
            Nodes.Free_Special(Nodes.Node_Class_Ptr(Flow));
            declare
            begin
               Destroy_Node(Nodes.Node_Class_Ptr(Flow));
            exception
               when Ex : others =>
                  Put_Line(Warning, "[Flow_Types] " & Ada.Exceptions.Exception_Name(Ex)  & ": Deallocation of a Node failed.");
            end;
         else
            Put_Line(Warning, "[Flow_Types] Unknown Object. Deallocation failed.");
         end if;
      end if;
   end Free;

   procedure Clear (TDescr: in out Type_Descr) is
   begin
      TDescr.Type_Name := S2u("");
      UStr_List.Clear(TDescr.Super_Type);
      Nodes.Free_Tree_Recursive(Nodes.Node_Class_Ptr(TDescr.Descr_Tree));
      UStr_List.Clear(TDescr.Component_Type);
      UStr_List_List.Free(TDescr.Used_Idents);
      Declaration_Ptr_List.Free(TDescr.Components);
      Declaration_Ptr_List.Free(TDescr.Discriminants);
      TDescr.Is_Private := False;
   end Clear;

   procedure Free (TDescr: in out Type_Descr_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Type_Descr, Type_Descr_Ptr);
   begin
      Clear(TDescr.all);
      Destroy(TDescr);
   end Free;

   procedure Free (Decl: in out Declaration_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Declaration, Declaration_Ptr);
   begin
      UStr_List.Clear(Decl.Vars);
      Clear(Decl.VType);
      Nodes.Free_Tree_Recursive(Nodes.Node_Class_Ptr(Decl.Init));
      Destroy(Decl);
   end Free;

   procedure Free (Alias: in out Alias_Class_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Flow_Types.Alias'Class, Alias_Class_Ptr);
   begin
      if Alias.all in Instance_Alias'Class then
         Alias_Class_Ptr_List.Free(Instance_Alias_Class_Ptr(Alias).Assocs);
      end if;
      UStr_List.Clear(Alias.Old_Name);
      Nodes.Free_Tree_Recursive(Nodes.Node_Class_Ptr(Alias.PTree));
      Destroy(Alias);
   end Free;

   procedure Free (FDecl: in out Formal_Decl_Class_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Formal_Decl'Class, Formal_Decl_Class_Ptr);
   begin
      UStr_List.Clear(FDecl.Vars);
      Clear(FDecl.Vtype);
      Nodes.Free_Tree_Recursive(Nodes.Node_Class_Ptr(FDecl.Init));
      Alias_Class_Ptr_List.Free(FDecl.Assocs);
      Destroy(FDecl);
   end Free;

   procedure Free (Ref: in out Ref_Class_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Reference'Class, Ref_Class_Ptr);
   begin
      UStr_List.Clear(Ref.Name);
      Destroy(Ref);
   end Free;

   function Get_Id (Flow: in Flow_Object) return Id.Pointer is
   begin
      return Flow.Oid;
   end Get_Id;

   procedure Set_Name (Flow: in out Flow_Object; Name: in Unbounded_String) is
   begin
      Flow.Name := Name;
   end Set_Name;

   function Get_Name (Flow: in Flow_Object) return String is
   begin
      return To_String(Flow.Name);
   end Get_Name;

   function Get_Element (Flow: in Flow_Object) return Asis.Element is
   begin
      return Flow.Element;
   end Get_Element;

   procedure Set_PTree (Flow: in out Flow_Object; PTree: in Flow_Class_Ptr) is
   begin
      Flow.PTree := PTree;
   end Set_PTree;

   function Get_PTree (Flow: in Flow_Object) return Flow_Class_Ptr is
   begin
      return Flow.PTree;
   end Get_PTree;

   procedure Add_Pred (Flow: in out Flow_Object; Pred: in Flow_Class_Ptr) is
   begin
      Flow_Class_Ptr_List.Append(Flow.Preds, Pred);
   end Add_Pred;

   procedure Rem_Pred (Flow: in out Flow_Object; Pred: in Flow_Class_Ptr) is
   begin
      Flow_Class_Ptr_List.Remove(Flow.Preds, Pred);
   end Rem_Pred;

   function Get_Preds (Flow: in Flow_Object) return Flow_Class_Ptr_List.Object is
   begin
      return Flow.Preds.all;
   end Get_Preds;

   procedure Add_Succ (Flow: in out Flow_Object; Succ: in Flow_Class_Ptr) is
   begin
      Flow_Class_Ptr_List.Append(Flow.Succs, Succ);
   end Add_Succ;

   procedure Rem_Succ (Flow: in out Flow_Object; Succ: in Flow_Class_Ptr) is
   begin
      Flow_Class_Ptr_List.Remove(Flow.Succs, Succ);
   end Rem_Succ;

   function Get_Succs (Flow: in Flow_Object) return Flow_Class_Ptr_List.Object is
   begin
      return Flow.Succs.all;
   end Get_Succs;

   procedure Clear_Succs (Flow: in out Flow_Object) is
   begin
      Flow_Class_Ptr_List.Clear(Flow.Succs);
   end Clear_Succs;

   procedure Rem_Last_Succ (Flow: in out Flow_Object) is
   begin
      Flow_Class_Ptr_List.Remove_Last (Flow.Succs);
   end Rem_Last_Succ;

   procedure Rem_Last_Pred (Flow: in out Flow_Object) is
   begin
      Flow_Class_Ptr_List.Remove_Last (Flow.Preds);
   end Rem_Last_Pred;

   function Get_Last_Pred (Flow: in Flow_Object) return Flow_Class_Ptr is
   begin
      if not Flow_Class_Ptr_List.Empty(Flow.Preds) then
         return Flow_Class_Ptr_List.Get_Tail(Flow.Preds);
      else
         return null;
      end if;
   end Get_Last_Pred;

   procedure Rem_All_Succs (Flow: in out Flow_Object) is
   begin
      Flow_Class_Ptr_List.Clear(Flow.Succs);
   end Rem_All_Succs;

   procedure Add_Decl (Flow: in out Flow_Object; Decl: in Declaration_Ptr) is
   begin
      Declaration_Ptr_List.Append(Flow.Decls, Decl);
   end Add_Decl;

   function Get_Decls (Flow: in Flow_Object) return Declaration_Ptr_List.Object is
   begin
      return Flow.Decls.all;
   end Get_Decls;

   procedure Add_FDecl (Flow: in out Flow_Object; FDecl: in Formal_Decl_Class_Ptr) is
   begin
      Formal_Decl_Class_Ptr_List.Append(Flow.FDecls, FDecl);
   end Add_FDecl;

   procedure Set_FDecls (Flow: in out Flow_Object; FDecls: in out Formal_Decl_Class_Ptr_List.Object) is
   begin
      Formal_Decl_Class_Ptr_List.Duplicate(FDecls, Flow.FDecls.all);
   end Set_FDecls;

   function Get_FDecls (Flow: in Flow_Object) return Formal_Decl_Class_Ptr_List.Object is
   begin
      return Flow.FDecls.all;
   end Get_FDecls;

   procedure Add_Ren (Flow: in out Flow_Object; Ren: in Alias_Class_Ptr) is
   begin
      Alias_Class_Ptr_List.Append(Flow.Rens, Ren);
   end Add_Ren;

   function Get_Rens (Flow: in Flow_Object) return Alias_Class_Ptr_List.Object is
   begin
      return Flow.Rens.all;
   end Get_Rens;

   procedure Add_Type (Flow: in out Flow_Object; TDescr: in Type_Descr_Ptr) is
   begin
      Type_Descr_Ptr_List.Append(Flow.Types, TDescr);
   end Add_Type;

   function Get_Types (Flow: in Flow_Object) return Type_Descr_Ptr_List.Object is
   begin
      return Flow.Types.all;
   end Get_Types;

   not overriding procedure Add_Ref (Flow: in out Flow_Object;
                                     Ref: in Ref_Class_Ptr) is
   begin
      Ref_Class_Ptr_List.Append(Flow.Refs, Ref);
   end Add_Ref;

   not overriding procedure Set_Refs (Flow: in out Flow_Object;
                                      Refs: in Ref_Class_Ptr_List.Object) is
     Local_Refs :Ref_Class_Ptr_List.Object := Refs;
   begin
      Ref_Class_Ptr_List.Duplicate(Local_Refs, Flow.Refs.all);
   end Set_Refs;

   not overriding function Get_Refs (Flow: in Flow_Object)
                                    return Ref_Class_Ptr_List.Object is
   begin
      return Flow.Refs.all;
   end Get_Refs;

end Ast2Cfg.Flow_Types;
