---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

package body Ast2Cfg.Generic_Pointer_List is

   function "=" (Left,Right: in Object) return Boolean is
   begin
      return List."="(Left.Lst, Right.Lst);
   end "=";

   function "=" (Left,Right: in Pointer) return Boolean is
   begin
      if Ptr_Equal(Left, null) and Ptr_Equal(Right, null) then
         return True;
      elsif Ptr_Equal(Left, null) or Ptr_Equal(Right, null) then
         return False;
      end if;

      return "="(Left.all, Right.all);
   end "=";

   procedure Init (Obj: out Object) is
   begin
      List.Init(Obj.Lst);
   end Init;

   procedure Init (Ptr: out Pointer) is
   begin
      if Ptr_Equal(Ptr, null) then
         Ptr := new Object;
      end if;
      Init(Ptr.all);
   end Init;

   procedure Append (Obj: in out Object; I: in Item_Pointer) is
   begin
      List.Append(Obj.Lst, I);
   end Append;

   procedure Append (Ptr: in out not null Pointer; I: in Item_Pointer) is
   begin
      Append(Ptr.all, I);
   end Append;

   procedure Append (Obj: in out Object; O: in out Object) is
   begin
      List.Append(Obj.Lst, O.Lst);
   end Append;

   procedure Append (Ptr: in out not null Pointer; P: in out not null Pointer) is
   begin
      Append(Ptr.all, P.all);
   end Append;

   procedure Insert_Head (Obj: in out Object; I: in Item_Pointer) is
   begin
      List.Insert_Head(Obj.Lst, I);
   end Insert_Head;

   procedure Insert_Head (Ptr: in out not null Pointer; I: in Item_Pointer) is
   begin
      Insert_Head(Ptr.all, I);
   end Insert_Head;

   function Get_Head (Obj: in Object) return Item_Pointer is
   begin
      return List.Get_Head(Obj.Lst);
   end Get_Head;

   function Get_Head (Ptr: in not null Pointer) return Item_Pointer is
   begin
      return Get_Head(Ptr.all);
   end Get_Head;

   function Get_Tail (Obj: in Object) return Item_Pointer is
   begin
      return List.Get_Tail(Obj.Lst);
   end Get_Tail;

   function Get_Tail (Ptr: in not null Pointer) return Item_Pointer is
   begin
      return Get_Tail(Ptr.all);
   end Get_Tail;

   procedure Replace_Head (Obj: in out Object; I: in Item_Pointer) is
   begin
      List.Replace_Head(Obj.Lst, I);
   end Replace_Head;

   procedure Replace_Head (Ptr: in out not null Pointer; I: in Item_Pointer) is
   begin
      Replace_Head(Ptr.all, I);
   end Replace_Head;

   procedure Remove_Last (Obj: in out Object) is
   begin
      List.Remove_Last(Obj.Lst);
   end Remove_Last;

   procedure Remove_Last (Ptr: in out not null Pointer) is
   begin
      Remove_Last(Ptr.all);
   end Remove_Last;

   procedure Remove (Obj: in out Object; I: in Item_Pointer) is
   begin
      List.Remove(Obj.Lst, I);
   end Remove;

   procedure Remove (Ptr: in out not null Pointer; I: in Item_Pointer) is
   begin
      Remove(Ptr.all, I);
   end Remove;

   procedure Duplicate (Src: in out Object; Dst: out Object) is
   begin
      List.Duplicate(Src.Lst, Dst.Lst);
   end Duplicate;

   procedure Duplicate (Src: in out not null Pointer; Dst: out Pointer) is
   begin
      Duplicate(Src.all, Dst.all);
   end Duplicate;

   procedure Clear (Obj: in out Object) is
   begin
      List.Clear(Obj.Lst);
   end Clear;

   procedure Clear (Ptr: in out not null Pointer) is
   begin
      Clear(Ptr.all);
   end Clear;

   procedure Free (Obj: in out Object) is
      Cur_Item_Pointer: Item_Pointer;
   begin
      while not List.Empty(Obj.Lst) loop
         List.Pop(Obj.Lst, Cur_Item_Pointer);
         Free_Procedure(Cur_Item_Pointer);
      end loop;
      List.Clear(Obj.Lst); -- to be on the safe side
   end Free;

   procedure Free (Ptr: in out Pointer) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Object, Pointer);
   begin
      if Ptr = null then
         return;
      end if;

      Free(Ptr.all);
      Destroy(Ptr);
   end Free;

   procedure Push (Obj: in out Object; I: in Item_Pointer) is
   begin
      List.Push(Obj.Lst, I);
   end Push;

   procedure Push (Ptr: in out not null Pointer; I: in Item_Pointer) is
   begin
      Push(Ptr.all, I);
   end Push;

   procedure Pop (Obj: in out Object; I: out Item_Pointer) is
   begin
      List.Pop(Obj.Lst, I);
   end Pop;

   procedure Pop (Ptr: in out not null Pointer; I: out Item_Pointer) is
   begin
      Pop(Ptr.all, I);
   end Pop;

   procedure Peek (Obj: in out Object; I: out Item_Pointer) is
   begin
      List.Peek(Obj.Lst, I);
   end Peek;

   procedure Peek (Ptr: in out not null Pointer; I: out Item_Pointer) is
   begin
      Peek(Ptr.all, I);
   end Peek;

   function Empty (Obj: in Object) return Boolean is
   begin
      return List.Empty(Obj.Lst);
   end Empty;

   function Empty (Ptr: in not null Pointer) return Boolean is
   begin
      return Empty(Ptr.all);
   end Empty;

   procedure Reset (Obj: in out Object) is
   begin
      List.Reset(Obj.Lst);
   end Reset;

   procedure Reset (Ptr: in out not null Pointer) is
   begin
      Reset(Ptr.all);
   end Reset;

   function Has_Next (Obj: in Object) return Boolean is
   begin
      return List.Has_Next(Obj.Lst);
   end Has_Next;

   function Has_Next (Ptr: in not null Pointer) return Boolean is
   begin
      return Has_Next(Ptr.all);
   end Has_Next;

   procedure Get_Next (Obj: in out Object; Next_Item_Pointer: out Item_Pointer) is
   begin
      List.Get_Next(Obj.Lst, Next_Item_Pointer);
   end Get_Next;

   procedure Get_Next (Ptr: in out not null Pointer; Next_Item_Pointer: out Item_Pointer) is
   begin
      Get_Next(Ptr.all, Next_Item_Pointer);
   end Get_Next;

   function Count (Obj: in Object) return Long_Long_Natural is
   begin
      return List.Count(Obj.Lst);
   end Count;

   function Count (Ptr: in not null Pointer) return Long_Long_Natural is
   begin
      return Count(Ptr.all);
   end Count;

end Ast2Cfg.Generic_Pointer_List;
