---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ast2Cfg.Utils; use Ast2Cfg.Utils;
with Ast2Cfg.Id;
with Ast2Cfg.Cfgs;

with Ast2Cfg.Output; use Ast2Cfg.Output;

with Asis;
with Asis.Elements;

package body Ast2Cfg.Nodes is

   overriding
   procedure Init (Node: in out Node_Object;
                   Oid:  in Id.Pointer;
                   Name: in Unbounded_String;
                   Pred: in Flow_Class_Ptr := null;
                   Element: in Asis.Element := Asis.Nil_Element) is -- private
   begin
      Flow_Types.Init(Flow_Object(Node), Oid, S2u(""), null, Element);
      Node.Info := S2u("");
      Node.Mark := 0;
      Node.RTree := null;
      UStr_List_List.Init(Node.Idents);
   end Init;

   not overriding
   procedure Init(Node: in out Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is
   begin
      Init(Node, Id.New_Node_Id, S2u(""), null, Element);
      Node.Graph := Graph;

      if Graph /= null then
         if Graph.all in Cfgs.CFG_Object'Class then
            Cfgs.Increment_Node_Sum(Cfgs.CFG_Object(Graph.all)); -- this is ugly, however, Graph has to be
         end if;                                                 -- Flow_Class_Ptr in order to avoid chicken & egg
      end if;
   end Init;


   not overriding
   procedure Free (Node: in out Node_Class_Ptr) is
   begin
      if Node /= null then
         Flow_Types.Free(Flow_Class_Ptr(Node));
      end if;
   end Free;

   not overriding
   procedure Free_Special (Node: in out Node_Class_Ptr) is
   begin
      if Node /= null then
         Free_Tree_Recursive(Node.RTree);
         UStr_List_List.Free(Node.Idents);
         -- I know this is ugly, however, Ada.Finalization turned out to be
         -- unusable with this program
         if Node.all in Call_Node_Object'Class then
            UStr_List.Clear(Call_Node_Object(Node.all).Dest_Name);
         elsif Node.all in Param_Node_Object'Class then
            UStr_List.Clear(Param_Node_Object(Node.all).Var_Name);
         elsif Node.all in Param_Alloc_Node_Object'Class then
            UStr_List.Clear(Param_Alloc_Node_Object(Node.all).Type_Name);
         elsif Node.all in Assign_Node_Object then
            Free_Tree_Recursive(Assign_Node_Object(Node.all).LTree);
         end if;
      end if;

      if Node.Graph /= null then
         if Node.Graph.all in Cfgs.CFG_Object'Class and
            Node.all not in Param_Node_Object'Class and
            Node.all not in Param_Alloc_Node_Object'Class and
            Node.all not in Param_Call_Node_Object'Class then
               Cfgs.Decrement_Node_Sum(Cfgs.CFG_Object(Node.Graph.all)); -- we counted it in the above Init, but now don't want to.
         end if;
      end if;

      if (Node.Graph /= null) and then
         (Node.Graph.all in CFGs.CFG_Object'Class) and then
         (Node = Cfgs.Get_Tail(Cfgs.CFG_Object(Node.Graph.all))) then
         Cfgs.Set_Tail(Cfgs.CFG_Object(Node.Graph.all), null);
      end if;
   end Free_Special;

   not overriding
   procedure Free_Tree_Recursive (Node: in out Node_Class_Ptr) is
     Succs: Flow_Class_Ptr_List.Object;
     Succ: Node_Class_Ptr := null;
   begin
      if Node = null then
         return;
      end if;
      Succs := Get_Succs(Node.all);
      while not Flow_Class_Ptr_List.Empty(Succs) loop
         Flow_Class_Ptr_List.Pop(Succs, Flow_Class_Ptr(Succ));
         Free_Tree_Recursive(Succ);
      end loop;
      Free(Node);
   end Free_Tree_Recursive;

   not overriding
   procedure Free_Recursive (Node: in out Node_Class_Ptr) is
     procedure Recursion (Free_List: in out Node_Class_Ptr_List.Pointer; Node: in out Node_Class_Ptr) is
        Succs: Flow_Class_Ptr_List.Object;
        Succ: Node_Class_Ptr := null;
     begin
        if Node = null or else Get_Mark(Node.all) = 4 then
           return;
        end if;
        Set_Mark(Node.all, 4);
        Node_Class_Ptr_List.Push(Free_List.all, Node);
        Succs := Get_Succs(Node.all);
        while not Flow_Class_Ptr_List.Empty(Succs) loop
           Flow_Class_Ptr_List.Pop(Succs, Flow_Class_Ptr(Succ));
           Recursion(Free_List, Succ);
        end loop;
     end Recursion;
     Free_List: Node_Class_Ptr_List.Pointer := new Node_Class_Ptr_List.Object;
   begin
      Node_Class_Ptr_List.Init(Free_List.all);
      Recursion(Free_List, Node);
      Node_Class_Ptr_List.Free(Free_List);
   end Free_Recursive;

   not overriding
   procedure Add_Ident (Node: in out Node_Object; Identifier: in out UStr_List.Object) is
      Ident_Copy: UStr_List.Object;
   begin
      UStr_List.Duplicate(Identifier, Ident_Copy);
      UStr_List_List.Append(Node.Idents, Ident_Copy);
   end Add_Ident;

   not overriding
   procedure Add_Idents (Node: in out Node_Object; Identifiers: in out UStr_List.Object) is
      Name_List: UStr_List.Object;
      Identifier: Unbounded_String := S2u("");
   begin
      UStr_List.Reset(Identifiers);
      while UStr_List.Has_Next(Identifiers) loop
         UStr_List.Get_Next(Identifiers, Identifier);
         Name_List := Tokenize(U2s(Identifier), ".");
         Add_Ident(Node, Name_List);
         UStr_List.Clear(Name_List);
      end loop;
   end Add_Idents;

   not overriding
   function Get_Idents (Node: in Node_Object) return UStr_List_List.Object is
   begin
      return Node.Idents;
   end Get_Idents;

   not overriding
   procedure Set_Info (Node: in out Node_Object; Info: Unbounded_String) is
   begin
      Node.Info := Info;
   end Set_Info;

   not overriding
   procedure Add_Info (Node: in out Node_Object; Info: Unbounded_String) is
   begin
      Node.Info := Node.Info & Info;
   end Add_Info;

   not overriding
   function Get_Info (Node: in Node_Object) return String is
   begin
      return To_String (Node.Info);
   end Get_Info;

   not overriding
   procedure Append_Info (Node: in out Node_Object; Info: Unbounded_String) is
   begin
      Ada.Strings.Unbounded.Append (Node.Info, Info);
   end Append_Info;

   -- Multi-Purpose Mark: Integer range 0..8; -1 for reset
   not overriding
   procedure Set_Mark (Node: in out Node_Object; Mark: in Node_Mark) is
   begin
      Node.Mark := Mark;
   end Set_Mark;

   not overriding
   function Get_Mark (Node: in Node_Object) return Node_Mark is
   begin
      return Node.Mark;
   end Get_Mark;

   -- resets the marks of all reachable nodes to 0 temporarily using -1
   not overriding
   procedure Reset_Mark (Node: in out Node_Class_Ptr) is
     procedure Recursion (List: in out Node_Class_Ptr_List.Pointer; Node: in out Node_Class_Ptr) is
        Succs: Flow_Class_Ptr_List.Object;
        Succ: Node_Class_Ptr := null;
     begin
        if Node = null or else Get_Mark(Node.all) = -1 then
           return;
        end if;
        Set_Mark(Node.all, -1);
        Node_Class_Ptr_List.Push(List.all, Node);
        Succs := Get_Succs(Node.all);

        Flow_Class_Ptr_List.Reset(Succs);
        while Flow_Class_Ptr_List.Has_Next(Succs) loop
           Flow_Class_Ptr_List.Get_Next(Succs, Flow_Class_Ptr(Succ));
           Recursion(List, Succ);
        end loop;
     end Recursion;

     List: Node_Class_Ptr_List.Pointer := new Node_Class_Ptr_List.Object;
     Cur_Node: Node_Class_Ptr := null;
   begin
      Node_Class_Ptr_List.Init(List.all);
      Recursion(List, Node);

      -- set all marks to 0
      Node_Class_Ptr_List.Reset(List);
      while Node_Class_Ptr_List.Has_Next(List) loop
         Node_Class_Ptr_List.Get_Next(List, Cur_Node);
         Set_Mark(Cur_Node.all, 0);
      end loop;
   end Reset_Mark;

   -- RTree

   not overriding
   procedure Set_RHS_Tree (Node: in out Node_Object; Tree: in Node_Class_Ptr) is
   begin
      Node.RTree := Tree;
   end Set_RHS_Tree;

   not overriding
   function Get_RHS_Tree (Node: in Node_Object) return Node_Class_Ptr is
   begin
      return Node.RTree;
   end Get_RHS_Tree;

   -- Call Node

   overriding
   procedure Init(Node: in out Call_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is
   begin
      Init (Node_Object(Node), Graph, Element);
      UStr_List.Init (Node.Dest_Name);
      Node.Dest_Ptr := null;
      Node.Dest_Elem := Asis.Nil_Element;
   end Init;

   procedure Set_Call_Target(Node: in out Call_Node_Object; Element: in Asis.Element) is
   begin
      Node.Dest_Elem := Element;
   end Set_Call_Target;

   function Get_Call_Target(Node: in Call_Node_Object) return Asis.Element is
   begin
      return Node.Dest_Elem;
   end Get_Call_Target;

   procedure Set_Call_Destination(Node: in out Call_Node_Object; Flow: Flow_Class_Ptr) is
   begin
      Node.Dest_Ptr := Flow;
      Put_Line(Debug,"[Nodes] Setting called flow <" & Flow_Types.Get_Name(Flow.all) & "> for call node <" & Nodes.Get_Dest_Name(Node) & ">");
   end Set_Call_Destination;

   function Get_Call_Destination(Node: in Call_Node_Object) return Flow_Class_Ptr is
   begin
      return Node.Dest_Ptr;
   end Get_Call_Destination;

   -- Add a Call Item to the Node (e.g. Ada + Text_IO + Put)
   not overriding
   procedure Add_Call_Item (Node: in out Call_Node_Object; Name: in Unbounded_String) is
      Name_List: UStr_List.Object;
      Component_Name: Unbounded_String := S2u("");
   begin
      Name_List := Tokenize(U2s(Name), ".");
      UStr_List.Reset(Name_List);
      while UStr_List.Has_Next(Name_List) loop
         UStr_List.Get_Next(Name_List, Component_Name);
         UStr_List.Append(Node.Dest_Name, Component_Name);
      end loop;
      UStr_List.Clear(Name_List);
   end Add_Call_Item;

   -- Get the complete Name of a Call
   not overriding
   function Get_Dest_Name (Node: in Call_Node_Object) return String is
     Call_Name, Item_Name: Unbounded_String := S2u("");
     Dest_Name: UStr_List.Object := Node.Dest_Name;
   begin
      UStr_List.Reset(Dest_Name);
      while (UStr_List.Has_Next(Dest_Name)) loop
         UStr_List.Get_Next(Dest_Name, Item_Name);
         if Call_Name /= "" then
            Call_Name := Call_Name & S2u(".");
         end if;
         Call_Name := Call_Name & Item_Name;
      end loop;
      return U2s(Call_Name);
   end Get_Dest_Name;

   -- Parameter Call Node

   overriding
   procedure Init(Node: in out Param_Call_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is
   begin
      Init (Node_Object(Node), Graph, Element);
      UStr_List.Init (Node.Dest_Name);
      Node.Dest_Ptr := null;

      if Graph /= null then
         if Graph.all in Cfgs.CFG_Object'Class then
            Cfgs.Decrement_Node_Sum(Cfgs.CFG_Object(Graph.all)); -- we counted it in the above Init, but didn't want to.
         end if;
      end if;
   end Init;

   -- Jump Node
   overriding
   procedure Init(Node: in out Jump_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is
   begin
      Init (Node_Object(Node), Graph, Element);
      Node.Target := S2u("");
   end Init;

   not overriding
   procedure Set_Target (Node: in out Jump_Node_Object; Target: in Unbounded_String) is
   begin
      Node.Target := Target;
   end Set_Target;

   not overriding
   function Get_Target (Node: in Jump_Node_Object) return String is
   begin
      return To_String(Node.Target);
   end Get_Target;

   -- Parameter Node
   overriding
   procedure Init (Node: in out Param_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is --private
   begin
      Init (Node_Object(Node), Graph, Element);
      UStr_List.Init(Node.Var_Name);

      if Graph /= null then
         if Graph.all in Cfgs.CFG_Object'Class then
            Cfgs.Decrement_Node_Sum(Cfgs.CFG_Object(Graph.all)); -- we counted it in the above Init, but didn't want to.
         end if;
      end if;
   end Init;

   not overriding
   procedure Init (Node: in out Param_Node_Object; Graph: in Flow_Class_Ptr := null) is
   begin
      Init(Node, Graph, Asis.Nil_Element);
   end Init;

   -- Parameter Allocation Node
   overriding
   procedure Init (Node: in out Param_Alloc_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is --private
   begin
      Init (Node_Object(Node), Graph, Element);
      UStr_List.Init(Node.Type_Name);

      if Graph /= null then
         if Graph.all in Cfgs.CFG_Object'Class then
            Cfgs.Decrement_Node_Sum(Cfgs.CFG_Object(Graph.all)); -- we counted it in the above Init, but didn't want to.
         end if;
      end if;
   end Init;

   not overriding
   procedure Init (Node: in out Param_Alloc_Node_Object; Graph: in Flow_Class_Ptr := null) is
   begin
      Init(Node, Graph, Asis.Nil_Element);
   end Init;

   -- Add a Param Item to the Node (e.g. Pkg + CFG + Var)
   not overriding
   procedure Add_Var_Item (Node: in out Param_Node_Object; Name: in Unbounded_String) is
      Name_List: UStr_List.Object;
      Component_Name: Unbounded_String := S2u("");
   begin
      Name_List := Tokenize(U2s(Name), ".");
      UStr_List.Reset(Name_List);
      while UStr_List.Has_Next(Name_List) loop
         UStr_List.Get_Next(Name_List, Component_Name);
         UStr_List.Append(Node.Var_Name, Component_Name);
      end loop;
      UStr_List.Clear(Name_List);
   end Add_Var_Item;

   -- Get the complete Name of the Variable
   not overriding
   function Get_Var_Name (Node: in Param_Node_Object) return String is
   begin
      return Merge_UStr_List(Node.Var_Name, ".");
   end Get_Var_Name;

   not overriding
   procedure Clear_Var (Node: in out Param_Node_Object) is
   begin
      UStr_List.Clear(Node.Var_Name);
   end Clear_Var;

   -- Add a Param Item to the Node (e.g. Pkg + CFG + Type)
   not overriding
   procedure Add_Type_Item (Node: in out Param_Alloc_Node_Object; Name: in Unbounded_String) is
      Name_List: UStr_List.Object;
      Component_Name: Unbounded_String := S2u("");
   begin
      Name_List := Tokenize(U2s(Name), ".");
      UStr_List.Reset(Name_List);
      while UStr_List.Has_Next(Name_List) loop
         UStr_List.Get_Next(Name_List, Component_Name);
         UStr_List.Append(Node.Type_Name, Component_Name);
      end loop;
      UStr_List.Clear(Name_List);
   end Add_Type_Item;

   -- Get the complete Name of the Type
   not overriding
   function Get_Type_Name (Node: in Param_Alloc_Node_Object) return String is
   begin
      return Merge_UStr_List(Node.Type_Name, ".");
   end Get_Type_Name;

   not overriding
   procedure Clear_Type (Node: in out Param_Alloc_Node_Object) is
   begin
      UStr_List.Clear(Node.Type_Name);
   end Clear_Type;

   -- Assign Node

   overriding
   procedure Init (Node: in out Assign_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element) is
   begin
      Init (Node_Object(Node), Graph, Element);
      Node.LTree := null;
   end Init;

   not overriding
   procedure Set_LHS_Tree (Node: in out Assign_Node_Object; Tree: in Node_Class_Ptr) is
   begin
      Node.LTree := Tree;
   end Set_LHS_Tree;

   not overriding
   function Get_LHS_Tree (Node: in Assign_Node_Object) return Node_Class_Ptr is
   begin
      return Node.LTree;
   end Get_LHS_Tree;

end Ast2Cfg.Nodes;
